/*
    Project: TCTEMP
    Version: V1.5
    File   : na_rng.c

    Copyright (c) 2006-2007 Author of TCTEMP. All rights reserved.
    Public Key Fingerprint: 75EB 6BC2 01B7 F6E7 4BD7 CC58 4A5F C393 19EE 6E69

    Covered by TrueCrypt License 2.2 the full text of which is contained
    in the file License.txt included in TCTEMP binary and source code
    distribution archives.

    The copyright notice of this source code file and the public key
    fingerprint must be retained in full.
*/

#include "na.h"

#define RNGT_NONE 0
#define RNGT_VIA  1

#define RNGR_UNSUCCESSFUL    -1
#define RNGR_NOT_IMPLEMENTED -2

#define GetNextAlignedPointer(p, a) \
  (PVOID) ( (((size_t)(p)) + ((a)-1)) & ~(size_t)((a)-1) )

#ifndef _WIN64

#define xstore __asm _emit 0x0f __asm _emit 0xa7 __asm _emit 0xc0

static int get_rng_type(void)
{
    int iRngType = RNGT_NONE;

    /* VIA */
    __asm {
        mov     eax, 0xc0000000
        cpuid
        cmp     eax, 0xc0000001
        jb      no_via_rng
        mov     eax, 0xc0000001
        cpuid
        and     edx, 0x0000000c
        cmp     edx, 0x0000000c
        jne     no_via_rng
        mov     iRngType, RNGT_VIA
    }
    no_via_rng:

    return iRngType;
}

#pragma warning(push)
#pragma warning(disable:4748)
static int get_rng_byte(int iRngType)
{
    unsigned __int8 rgbBuffer[32];
    unsigned __int8 *pbBuffer;
    int             iResult;

    pbBuffer = GetNextAlignedPointer(rgbBuffer, 16);

    if (iRngType == RNGT_VIA) {

        do {

            __asm {
                mov     edi, pbBuffer
                mov     edx, 3
                xstore
                and     eax, 0x0000000f
                jz      unsuccessful
            }

            iResult = *pbBuffer;
            break;

        unsuccessful:
            iResult = RNGR_UNSUCCESSFUL;

        } while (0);

    } else {

        iResult = RNGR_NOT_IMPLEMENTED;

    }

    return iResult;
}
#pragma warning(pop)

#else

#define get_rng_type()         RNGT_NONE
#define get_rng_byte(iRngType) RNGR_NOT_IMPLEMENTED

#endif

#ifndef get_rng_type

int get_rng_bytes(unsigned __int8 *pbBuffer, size_t cbBufferSize)
{
    int iResult = 0;

    __try {

        unsigned __int8 *pbBufferEnd;
        int             iRngType;

        iRngType = get_rng_type();
        if (iRngType == RNGT_NONE) {
            goto error;
        }

        pbBufferEnd = pbBuffer + cbBufferSize;

        while (pbBuffer != pbBufferEnd) {

            int i;

            i = get_rng_byte(iRngType);
            if (i == RNGR_UNSUCCESSFUL) {
                i = get_rng_byte(iRngType);
                if (i == RNGR_UNSUCCESSFUL) {
                    NaSleep(200);
                    i = get_rng_byte(iRngType);
                }
            }
            if (i < 0) {
                goto error;
            }

            *pbBuffer++ = (unsigned __int8) i;
        }

        iResult = 1;

    error:
        ;

    } __except(EXCEPTION_EXECUTE_HANDLER) {
    }

    return iResult;
}

#else

int get_rng_bytes(unsigned __int8 *pbBuffer, size_t cbBufferSize)
{
    return 0;
}

#endif
